@@ -3,7 +3,7 @@ package Crypt::JWT;
 use strict;
 use warnings;
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 use Exporter 'import';
 our %EXPORT_TAGS = ( all => [qw(decode_jwt encode_jwt)] );
@@ -69,6 +69,132 @@ sub _b64u_to_hash {
   return $hash;
 }
 
+sub _add_claims {
+  my ($payload, %args) = @_;
+  #### claims (defined for JWS only)
+  # "exp"   Expiration Time
+  # "nbf"   Not Before
+  # "iat"   Issued At
+  # "iss"   Issuer
+  # "sub"   Subject
+  # "aud"   Audience
+  # "jti"   JWT ID
+  my $now = time;
+  $payload->{iat} = $now                       if $args{auto_iat};
+  $payload->{exp} = $now + $args{relative_exp} if defined $args{relative_exp};
+  $payload->{nbf} = $now + $args{relative_nbf} if defined $args{relative_nbf};
+}
+
+sub _verify_claims {
+  my ($payload, %args) = @_;
+  
+  my $leeway = $args{leeway} || 0;
+  my $now = time;
+
+  ### iat
+  if(defined $payload->{iat}) {
+    if (!defined $args{verify_iat} || $args{verify_iat}==1) {
+      croak "JWT: iat claim check failed" if $payload->{iat} - $leeway < $now;
+    }
+  }
+  elsif ($args{verify_iat} && $args{verify_iat}==1) {
+    croak "JWT: iat claim required but missing"
+  }
+  
+  ### nbf
+  if(defined $payload->{nbf}) {
+    if (!defined $args{verify_nbf} || $args{verify_nbf}==1) {
+      croak "JWT: nbf claim check failed" if $payload->{nbf} - $leeway < $now;
+    }
+  }
+  elsif ($args{verify_nbf} && $args{verify_nbf}==1) {
+    croak "JWT: nbf claim required but missing"
+  }
+    
+  ### exp
+  if(defined $payload->{exp}) {
+    if (!defined $args{verify_exp} || $args{verify_exp}==1) {
+      croak "JWT: exp claim check failed" if $payload->{exp} + $leeway > $now;
+    }
+  }
+  elsif ($args{verify_exp} && $args{verify_exp}==1) {
+    croak "JWT: exp claim required but missing"
+  }
+
+  ### iss
+  if(exists $payload->{iss}) {
+    if (ref $args{verify_iss} eq 'Regexp') {
+      croak "JWT: iss claim re check failed" unless $payload->{iss} =~ $args{verify_iss};
+    }
+    elsif (ref $args{verify_iss} eq 'CODE') {
+      croak "JWT: iss claim check failed" unless $args{verify_iss}->($payload->{iss});
+    }
+  }
+  
+  ### sub
+  if(exists $payload->{sub}) {
+    if (ref $args{verify_sub} eq 'Regexp') {
+      croak "JWT: sub claim re check failed" unless $payload->{sub} =~ $args{verify_sub};
+    }
+    elsif (ref $args{verify_sub} eq 'CODE') {
+      croak "JWT: sub claim check failed" unless $args{verify_sub}->($payload->{sub});
+    }
+  }
+  
+  ### aud
+  if(exists $payload->{aud}) {
+    if (ref $args{verify_aud} eq 'Regexp') {
+      croak "JWT: aud claim re check failed" unless $payload->{aud} =~ $args{verify_aud};
+    }
+    elsif (ref $args{verify_aud} eq 'CODE') {
+      croak "JWT: aud claim check failed" unless $args{verify_aud}->($payload->{aud});
+    }
+  }
+  
+  ### jti
+  if(exists $payload->{jti}) {
+    if (ref $args{verify_jti} eq 'Regexp') {
+      croak "JWT: jti claim re check failed" unless $payload->{jti} =~ $args{verify_jti};
+    }
+    elsif (ref $args{verify_jti} eq 'CODE') {
+      croak "JWT: jti claim check failed" unless $args{verify_jti}->($payload->{jti});
+    }
+  }
+}
+
+sub _payload_zip {
+  my ($payload, $header, $z) = @_;
+  my @zip = ref $z eq 'ARRAY' ? @$z : ($z);
+  if ($zip[0] eq 'deflate') {
+    my $level = defined $zip[1] ? $zip[1] : 6;
+    $header->{zip} = "DEF";
+    my $d = Compress::Raw::Zlib::Deflate->new(-Bufsize => 300, -Level => $level, -WindowBits => -MAX_WBITS());
+    my $output = '';
+    $d->deflate($payload, $output);
+    croak "JWT: deflate failed" unless $output;
+    $payload = $output;
+  }
+  else {
+    croak "JWT: unknown zip method '$zip[0]'";
+  }
+  return $payload;
+}
+
+sub _payload_unzip {
+  my ($payload, $z) = @_;
+  if ($z eq "DEF") {
+    my $d = Compress::Raw::Zlib::Inflate->new(-Bufsize => 300, -WindowBits => -MAX_WBITS());
+    my $output = '';
+    $d->inflate($payload, $output);
+    croak "JWT: inflate failed" unless $output;
+    $payload = $output;
+  }
+  else {
+    croak "JWT: unknown zip method '$z'";
+  }
+  return $payload;
+}
+
 sub _payload_enc {
   my ($payload) = @_;
   if (ref($payload) =~ /^(HASH|ARRAY)$/) {
@@ -248,41 +374,29 @@ sub _encode_jwe {
   my $alg     = $args{alg};
   my $enc     = $args{enc};
   my $header  = $args{extra_headers} ? \%{$args{extra_headers}} : {};
-    # prepare header
-    $header->{alg} = $alg;
-    $header->{enc} = $enc;
-    # prepare cek
-    my ($cek, $ecek) = _encrypt_cek($key, $header);
-    # encode header
-    my $json_header = encode_json($header);
-    my $b64u_header = encode_base64url($json_header);
-    # add claims
-    if (ref $payload eq 'HASH') {
-      $payload->{iat} = time if $args{iat_auto};
-    }
-    # serialize payload
-    $payload = _payload_enc($payload);
-    # compress payload
-    if (my $z = $args{zip}) {
-      my @zip = ref $z eq 'ARRAY' ? @$z : ($z);
-      if ($zip[0] eq 'deflate') {
-        my $level = defined $zip[1] ? $zip[1] : 6;
-        $header->{zip} = "DEF";
-        my $d = Compress::Raw::Zlib::Deflate->new(-Bufsize => 300, -Level => $level, -WindowBits => -MAX_WBITS());
-        my $output = '';
-        $d->deflate($payload, $output);
-        croak "JWT: deflate failed" unless $output;
-        $payload = $output;
-      }
-    }
-    # encrypt payload
-    my ($ct, $iv, $tag) = _encrypt_jwe_payload($cek, $enc, $b64u_header, $payload);
-    # return token parts
-    return ( $b64u_header,
-             encode_base64url($ecek),
-             encode_base64url($iv),
-             encode_base64url($ct),
-             encode_base64url($tag));
+  # add claims to payload
+  _add_claims($payload, %args) if ref $payload eq 'HASH';
+  # serialize payload
+  $payload = _payload_enc($payload);
+  # compress payload
+  $payload = _payload_zip($payload, $header, $args{zip}) if $args{zip}; # may set some header items
+  # prepare header
+  $header->{alg} = $alg;
+  $header->{enc} = $enc;
+  $header->{typ} = 'JWE' if $args{auto_typ};
+  # prepare cek
+  my ($cek, $ecek) = _encrypt_cek($key, $header); # adds some header items
+  # encode header
+  my $json_header = encode_json($header);
+  my $b64u_header = encode_base64url($json_header);
+  # encrypt payload
+  my ($ct, $iv, $tag) = _encrypt_jwe_payload($cek, $enc, $b64u_header, $payload);
+  # return token parts
+  return ( $b64u_header,
+           encode_base64url($ecek),
+           encode_base64url($iv),
+           encode_base64url($ct),
+           encode_base64url($tag));
 }
 
 sub _decode_jwe {
@@ -314,19 +428,10 @@ sub _decode_jwe {
 
   my $cek = _decrypt_jwe_cek($ecek, $key, $header);
   my $payload = _decrypt_jwe_payload($cek, $header->{enc}, $b64u_header, $ct, $iv, $tag);
-  if (my $zip = $header->{zip}) {
-    if ($zip eq "DEF") {
-      my $d = Compress::Raw::Zlib::Inflate->new(-Bufsize => 300, -WindowBits => -MAX_WBITS());
-      my $output = '';
-      $d->inflate($payload, $output);
-      croak "JWT: inflate failed" unless $output;
-      $payload = $output;
-    }
-    else {
-      croak "JWT: unknown zip method '$zip'";
-    }
-  }
-  return ($header, _payload_dec($payload, $args{decode_payload}));
+  $payload = _payload_unzip($payload, $header->{zip}) if $header->{zip};
+  $payload = _payload_dec($payload, $args{decode_payload});
+  _verify_claims($payload, %args) if ref $payload eq 'HASH'; # croaks on error
+  return ($header, $payload);
 }
 
 sub _sign_jws {
@@ -387,18 +492,24 @@ sub _verify_jws {
 sub _encode_jws {
   my %args = @_;
   my $key     = $args{key};
-  my $payload = _payload_enc($args{payload});
+  my $payload = $args{payload};
   my $alg     = $args{alg};
   my $header  = $args{extra_headers} ? \%{$args{extra_headers}} : {};
   croak "JWS: alg 'none' not allowed" if $alg eq 'none' && !$args{allow_none};
+  # add claims to payload
+  _add_claims($payload, %args) if ref $payload eq 'HASH';
+  # serialize payload
+  $payload = _payload_enc($payload);
+  # compress payload
+  $payload = _payload_zip($payload, $header, $args{zip}) if $args{zip}; # may set some header items
+  # encode payload
+  my $b64u_payload = encode_base64url($payload);
   # prepare header
-  $header->{typ} = 'JWS' if $args{auto_typ};
   $header->{alg} = $alg;
+  $header->{typ} = 'JWS' if $args{auto_typ};
   # encode header
   my $json_header = encode_json($header);
   my $b64u_header = encode_base64url($json_header);
-  # encode payload
-  my $b64u_payload = encode_base64url($payload);
   # sign header
   my $b64u_signature = _sign_jws($b64u_header, $b64u_payload, $alg, $args{key}, $args{keypass});
   return ($b64u_header, $b64u_payload, $b64u_signature);
@@ -424,7 +535,10 @@ sub _decode_jws {
     croak "JWS: decode failed" if !$valid;
   }
   my $payload = decode_base64url($b64u_payload);
-  return ($header, _payload_dec($payload, $args{decode_payload}));
+  $payload = _payload_unzip($payload, $header->{zip}) if $header->{zip};
+  $payload = _payload_dec($payload, $args{decode_payload});
+  _verify_claims($payload, %args) if ref $payload eq 'HASH'; # croaks on error
+  return ($header, $payload);
 }
 
 sub encode_jwt {
@@ -491,34 +605,15 @@ sub decode_jwt {
 # "enc" (Encryption Algorithm) Header Parameter
 # "zip" (Compression Algorithm) Header Parameter
 
-#### claim (JWS only)
-# "iss"   Issuer
-# "sub"   Subject
-# "aud"   Audience
-# "exp"   Expiration Time
-# "nbf"   Not Before
-# "iat"   Issued At
-# "jti"   JWT ID
+#XXX-TODO more named args:
 
 #### encode:
+# - auto_jwk
 # - keys
-# - set_jwk       0|1
-# - set_iat       0|1
-# - set_exp       undef|<num>
-# - set_nbf       undef|<num>
-# - set_typ       0|1
-# - serializer compact|json
+# - serializer  compact|json
 #
 #### decode:
 # - keys
-# - leeway
-# - verify_iat        undef|0|1|sub
-# - verify_nbf        undef|0|1|sub
-# - verify_exp        undef=check_if_defined|0=ignore|1=require|sub
-# - verify_iss        undef|sub|regexp
-# - verify_aud        undef|sub|regexp
-# - verify_sub        undef|sub|regexp
-# - verify_jti        undef|sub|regexp
 
 =pod
 
@@ -558,7 +653,7 @@ Named arguments:
 
 =over
 
-=item token (mandatory)
+=item * token (mandatory)
 
 A string with either JWS or JWE JSON Web Token.
 
@@ -568,31 +663,31 @@ A string with either JWS or JWE JSON Web Token.
  ### JWE token example (5 segments)
  $t="eyJlbmMiOiJBMTI4R0NNIiwiYWxnIjoiQTEyOEtXIn0.UusxEbzhGkORxTRq0xkFKhvzPrXb9smw.VGfOuq0Fxt6TsdqLZUpnxw.JajIQQ.pkKZ7MHS0XjyGmRsqgom6w";
 
-=item key (mostly mandatory)
+=item * key (mostly mandatory)
 
 A key used for token decryption (JWE) or token signature validation (JWS). The value depends on C<alg> token header value.
 
- JWS alg header   key value
- --------------   ----------------
- none             no key required
- HS256            string (raw octects) of any length
- HS384            dtto
- HS512            dtto
- RS256            public RSA key, a string with PEM or DER or JSON/JWK key,
-                  perl hash with JWK key, an instance of Crypt::PK::RSA or
-                  Crypt::OpenSSL::RSA or Crypt::X509 or Crypt::OpenSSL::X509
- RS384            public RSA key, see RS256
- RS512            public RSA key, see RS256
- PS256            public RSA key, see RS256
- PS384            public RSA key, see RS256
- PS512            public RSA key, see RS256
- ES256            public ECC key, a string with PEM or DER or JSON/JWK key,
-                  perl hash with JWK key, an instance of Crypt::PK::ECC
- ES384            public ECC key, see ES256
- ES512            public ECC key, see ES256
+ JWS alg header      key value
+ ------------------  ----------------------------------
+ none                no key required
+ HS256               string (raw octects) of any length
+ HS384               dtto
+ HS512               dtto
+ RS256               public RSA key, a string with PEM or DER or JSON/JWK key,
+                     perl hash with JWK key, an instance of Crypt::PK::RSA or
+                     Crypt::OpenSSL::RSA or Crypt::X509 or Crypt::OpenSSL::X509
+ RS384               public RSA key, see RS256
+ RS512               public RSA key, see RS256
+ PS256               public RSA key, see RS256
+ PS384               public RSA key, see RS256
+ PS512               public RSA key, see RS256
+ ES256               public ECC key, a string with PEM or DER or JSON/JWK key,
+                     perl hash with JWK key, an instance of Crypt::PK::ECC
+ ES384               public ECC key, see ES256
+ ES512               public ECC key, see ES256
 
  JWE alg header      key value
- ------------------  ----------------
+ ------------------  ----------------------------------
  dir                 string (raw octects), length depends on 'enc' algorithm
  A128KW              string (raw octects) 16 bytes
  A192KW              string (raw octects) 24 bytes
@@ -614,23 +709,23 @@ A key used for token decryption (JWE) or token signature validation (JWS). The v
  ECDH-ES+A192KW      private ECC key, see ECDH-ES
  ECDH-ES+A256KW      private ECC key, see ECDH-ES
 
-=item keypass (optional, default: C<undef>)
+=item * keypass (optional, default: C<undef>)
 
 When 'key' parameter is an encrypted private RSA or ECC key this parameter may contain a password for private key decryption.
 
-=item allow_none (optional, default: C<0>)
+=item * allow_none (optional, default: C<0>)
 
 C<1> - accept JWS tokens with C<none> 'alg' header value (which means that token has no signature), B<BEWARE: DANGEROUS, UNSECURE!!!>
 
 C<0> - do not allow JWS with C<none> 'alg' header value
 
-=item ignore_signature (optional, default: C<0>)
+=item * ignore_signature (optional, default: C<0>)
 
 C<1> - do not check signature on JWS tokens, B<BEWARE: DANGEROUS, UNSECURE!!!>
 
 C<0> - check signature on JWS tokens
 
-=item accepted_alg (optional, default: C<undef>)
+=item * accepted_alg (optional, default: C<undef>)
 
 C<undef> means accept all 'alg' algorithms except 'none' (for accepting 'none' use C<allow_none>)
 
@@ -646,7 +741,7 @@ C<Regexp> that has to match 'alg' algorithm name
  #or
  my $payload = decode_jwt(token=>$t, key=>$k, accepted_enc=>qr/^HS(256|384|512)$/);
 
-=item accepted_enc (optional, default: C<undef>)
+=item * accepted_enc (optional, default: C<undef>)
 
 C<undef> means accept all 'enc' algorithms
 
@@ -662,7 +757,7 @@ C<Regexp> that has to match 'enc' algorithm name
  #or
  my $payload = decode_jwt(token=>$t, key=>$k, accepted_enc=>qr/^A(128|192|256)GCM$/);
 
-=item decode_payload (optional, default: C<undef>)
+=item * decode_payload (optional, default: C<undef>)
 
 C<0> - do not decode payload, return it as a raw string (octects).
 
@@ -670,7 +765,7 @@ C<1> - decode payload from JSON string, return it as perl hash ref (or array ref
 
 C<undef> - if possible decode payload from JSON string, if decode_json fails return payload as a raw string (octets).
 
-=item decode_header (optional, default: C<0>)
+=item * decode_header (optional, default: C<0>)
 
 C<0> - do not return decoded header as a return value of decode_jwt()
 
@@ -680,6 +775,58 @@ C<1> - return decoded header as a return value of decode_jwt()
  #or
  my ($header, $payload) = decode_jwt(token=>$t, key=>$k, decode_header=>1);
 
+=item * verify_iss (optional, default: C<undef>)
+
+C<CODE ref> - subroutine (with 'iss' claim value passed as argument) should return C<true> otherwise verification fails
+
+C<Regexp ref> - 'iss' claim value has to match given regexp otherwise verification fails
+
+=item * verify_aud (optional, default: C<undef>)
+
+C<CODE ref> - subroutine (with 'aud' claim value passed as argument) should return C<true> otherwise verification fails
+
+C<Regexp ref> - 'aud' claim value has to match given regexp otherwise verification fails
+
+=item * verify_sub (optional, default: C<undef>)
+
+C<CODE ref> - subroutine (with 'sub' claim value passed as argument) should return C<true> otherwise verification fails
+
+C<Regexp ref> - 'sub' claim value has to match given regexp otherwise verification fails
+
+=item * verify_jti (optional, default: C<undef>)
+
+C<CODE ref> - subroutine (with 'jti' claim value passed as argument) should return C<true> otherwise verification fails
+
+C<Regexp ref> - 'jti' claim value has to match given regexp otherwise verification fails
+
+=item * verify_iat (optional, default: C<undef>)
+
+C<0> - ignore 'iat' claim
+
+C<1> - require valid 'iat' claim
+
+C<undef> - 'iat' claim must be valid if present
+
+=item * verify_nbf (optional, default: C<undef>)
+
+C<0> - ignore 'nbf' claim
+
+C<1> - require valid 'nbf' claim
+
+C<undef> - 'nbf' claim must be valid if present
+
+=item * verify_exp (optional, default: C<undef>)
+
+C<0> - ignore 'exp' claim
+
+C<1> - require valid 'exp' claim
+
+C<undef> - 'exp' claim must be valid if present
+
+=item * leeway (optional, default: C<0>)
+
+Tolerance in seconds related to C<verify_exp>, C<verify_nbf>, C<verify_iat>.
+
 =back
 
 =head2 encode_jwt
@@ -755,27 +902,27 @@ Supported 'enc' algorithms:
 
 A key used for token encryption (JWE) or token signing (JWS). The value depends on C<alg> token header value.
 
- JWS alg header   key value
- --------------   ----------------
- none             no key required
- HS256            string (raw octects) of any length
- HS384            dtto
- HS512            dtto
- RS256            private RSA key, a string with PEM or DER or JSON/JWK key,
-                  perl hash with JWK key, an instance of Crypt::PK::RSA or
-                  Crypt::OpenSSL::RSA or Crypt::X509 or Crypt::OpenSSL::X509
- RS384            private RSA key, see RS256
- RS512            private RSA key, see RS256
- PS256            private RSA key, see RS256
- PS384            private RSA key, see RS256
- PS512            private RSA key, see RS256
- ES256            private ECC key, a string with PEM or DER or JSON/JWK key,
-                  perl hash with JWK key, an instance of Crypt::PK::ECC
- ES384            private ECC key, see ES256
- ES512            private ECC key, see ES256
+ JWS alg header      key value
+ ------------------  ----------------------------------
+ none                no key required
+ HS256               string (raw octects) of any length
+ HS384               dtto
+ HS512               dtto
+ RS256               private RSA key, a string with PEM or DER or JSON/JWK key,
+                     perl hash with JWK key, an instance of Crypt::PK::RSA or
+                     Crypt::OpenSSL::RSA or Crypt::X509 or Crypt::OpenSSL::X509
+ RS384               private RSA key, see RS256
+ RS512               private RSA key, see RS256
+ PS256               private RSA key, see RS256
+ PS384               private RSA key, see RS256
+ PS512               private RSA key, see RS256
+ ES256               private ECC key, a string with PEM or DER or JSON/JWK key,
+                     perl hash with JWK key, an instance of Crypt::PK::ECC
+ ES384               private ECC key, see ES256
+ ES512               private ECC key, see ES256
 
  JWE alg header      key value
- ------------------  ----------------
+ ------------------  ----------------------------------
  dir                 string (raw octects), length depends on 'enc' algorithm
  A128KW              string (raw octects) 16 bytes
  A192KW              string (raw octects) 24 bytes
@@ -819,6 +966,32 @@ Compression method, currently 'deflate' is the only one supported.
  #or define compression level
  my $token = encode_jwt(payload=>$p, key=>$k, alg=>'HS256', zip=>['deflate', 9]);
 
+=item auto_iat (optional, default: C<0>)
+
+C<1> - set 'iat' (Issued At) claim to current time (epoch seconds since 1970) at the moment of token encoding
+
+C<0> - do not set 'iat' claim
+
+NOTE: claims are part of the payload and can be used only if the payload is a HASH ref!
+
+=item relative_exp (optional, default: C<undef>)
+
+Set 'exp' claim (Expiration Time) to current time + C<relative_exp> value (in seconds).
+
+NOTE: claims are part of the payload and can be used only if the payload is a HASH ref!
+
+=item relative_nbf (optional, default: C<undef>)
+
+Set 'nbf' claim (Not Before) to current time + C<relative_exp> value (in seconds).
+
+NOTE: claims are part of the payload and can be used only if the payload is a HASH ref!
+
+=item auto_typ (optional, default: C<0>)
+
+C<1> - set 'typ' header to value C<JWS> for JWS tokens and C<JWE> for JWE tokens.
+
+C<0> - do not set 'typ' header
+
 =back
 
 =head1 SEE ALSO
